/*
 * config.c: Routines to read/write the vtxrc-file
 *
 * Copyright (c) 1994-96 Martin Buck  <martin-2.buck@student.uni-ulm.de>
 * Read COPYING for more information
 *
 */

#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <time.h>
#include <errno.h>
#include <string.h>
#include <sys/vtx.h>
#include "vtx_assert.h"
#include "safe_malloc.h"
#include "misc.h"
#include "xinit.h"
#include "config.h"


const scr_update_t update_table[] = {
  { "None", 3 },
  { "Slow", 2 },
  { "Fast", 0 }
};
const int update_table_count = 3;


typedef struct {
  char *spooldir;
  char *printer[PRN_COUNT];
  vtxdisp_t display_mode;
  int update_speed, font;
  int popup_rows[POPUP_COUNT - 1];			/* POPUP_TOPTEXT is unused here */
  placement_t popup_placement[POPUP_COUNT];
  int popup_startup[POPUP_COUNT];
  int decoration[WM_COUNT];
  unsigned tv_disp : 1;
  unsigned interlace : 1;
  unsigned sticky_fcs : 1;
  unsigned auto_top : 1;
  unsigned auto_reset : 1;
  unsigned cont_hot_search : 1;
  unsigned cont_search : 1;
  unsigned spool_only : 1;
} config_t;

static config_t conf;
static const char *placement_str[] = { "centered", "left", "right", "top", "bottom" };
static const char *bool_str[] = { "false", "true" };

void
config_init(void) {
  int printer, popup;
  
  config_set_spooldir("");
  for (printer = 0; printer < PRN_COUNT; printer++) {
    config_set_printer(printer, "|lpr");
  }
  config_set_display_mode(DISPOFF);
  config_set_cont_search(0);
  config_set_update_speed(1);
  for (popup = 0; popup < POPUP_COUNT; popup++) {
    if (popup < POPUP_TOPTEXT) {
      config_set_popup_rows(popup, popup == POPUP_SELST ? 5 : 9);
    }
    config_set_placement(popup, PLACEMENT_CENTER);
    config_set_startup(popup, PLACEMENT_CENTER);
  }
  config_set_tv_disp(FALSE);
  config_set_interlace(TRUE);
  config_set_sticky_fcs(FALSE);
  config_set_auto_top(FALSE);
  config_set_auto_reset(TRUE);
  config_set_cont_hot_search(TRUE);
  config_set_cont_search(FALSE);
  config_set_spool_only(FALSE);
  config_set_wmdeco(WM_OFS_X, 0);
  config_set_wmdeco(WM_OFS_Y, 0);
  config_set_wmdeco(WM_LB, 0);
  config_set_wmdeco(WM_RB, 0);
  config_set_wmdeco(WM_TB, 0);
  config_set_wmdeco(WM_BB, 0);
  config_set_wmdeco(WM_ADJUSTED, WM_ADJ_NONE);
}


int
config_parse_bool(const char *key, const char *val, const char *name, int *retval, int *invalid) {
  if (!name || !strcasecmp(key, name)) {
    if (!strcasecmp(val, "true") || !strcmp(val, "1")) {
      *retval = TRUE;
    } else if (!strcasecmp(val, "false") || !strcmp(val, "0")) {
      *retval = FALSE;
    } else {
      *invalid = TRUE;
      return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}


int
config_parse_int(const char *key, const char *val, const char *name, int min, int max,
    int base, int *retval, int *invalid) {
  char *tail;
  int tmpval;
  
  if (!name || !strcasecmp(key, name)) {
    tmpval = strtol(val, &tail, base);
    if (*tail == '\0' && tmpval >= min && tmpval <= max) {
      *retval = tmpval;
      return TRUE;
    } else {
      *invalid = TRUE;
      return FALSE;
    }
  }
  return FALSE;
}


int
config_load(const char *fname, int private) {
  int err, invalid, lcount = 0, popup_p, popup_s, tmpval;
  char line[256], token[100], tmpstr[100], sep[2];
  FILE *file;

  if (!(file = fopen(fname, "r"))) {
    return -1;
  }
  while (1) {
    if (!fgets(line, 256, file)) {
      if (feof(file))
        break;
      err = errno;
      fclose(file);
      errno = err;
      return -1;
    }
    lcount++;
    tmpstr[0] = '\0';
    popup_p = popup_s = -1;
    invalid = FALSE;
    
    if (line[0] == '#' || sscanf(line, " %c ", token) <= 0) {	/* Skip comments or empty lines */
      continue;
    } else if (sscanf(line, " %99[^\1- =] %1[=] %99[^\1-\37] ", token, sep, tmpstr) < 2) {
      fprintf(stderr, "%s:%d: Warning: Syntax error (ignoring line).\n", fname, lcount);
      continue;
    }
    if (!strcasecmp(token, "font")) {
      int font;
      
      for (font = 0; font <= vtxmaxfont; font++) {
        if (!strcasecmp(tmpstr, vtxfonts[font].fontname)) {
          config_set_font(font);
          break;
        }
      }
      if (font > vtxmaxfont) {
        invalid = TRUE;
      }
    } else if (!strcasecmp(token, "screen_updates")) {
      int speed;
      
      for (speed = 0; speed < update_table_count; speed++) {
        if (!strcasecmp(tmpstr, update_table[speed].menu_string)) {
          config_set_update_speed(speed);
          break;
        }
      }
      if (speed > 2) {
        invalid = TRUE;
      }
    } else if (!strcasecmp(token, "history_placement")) {
      popup_p = POPUP_HISTORY;
    } else if (!strcasecmp(token, "toptext_placement")) {
      popup_p = POPUP_TOPTEXT;
    } else if (!strcasecmp(token, "hotlist_placement")) {
      popup_p = POPUP_HOTLIST;
    } else if (!strcasecmp(token, "stationlist_placement")) {
      popup_p = POPUP_SELST;
    } else if (!strcasecmp(token, "history_startup")) {
      popup_s = POPUP_HISTORY;
    } else if (!strcasecmp(token, "toptext_startup")) {
      popup_s = POPUP_TOPTEXT;
    } else if (!strcasecmp(token, "hotlist_startup")) {
      popup_s = POPUP_HOTLIST;
    } else if (!strcasecmp(token, "stationlist_startup")) {
      popup_s = POPUP_SELST;
    } else if (!strcasecmp(token, "display_mode")) {
      if (!strcasecmp(tmpstr, "normal")) {
        config_set_display_mode(DISPNORM);
      } else if (!strcasecmp(tmpstr, "transparent")) {
        config_set_display_mode(DISPTRANS);
      } else if (!strcasecmp(tmpstr, "insert")) {
        config_set_display_mode(DISPINS);
      } else {
        invalid = TRUE;
      }
    } else if (!strcasecmp(token, "spooldir")) {
      config_set_spooldir(tmpstr);
    } else if (!strcasecmp(token, "default_cps_printer")) {
      config_set_printer(PRN_CPS, tmpstr);
    } else if (!strcasecmp(token, "default_ps_printer")) {
      config_set_printer(PRN_PS, tmpstr);
    } else if (!strcasecmp(token, "default_ascii_printer")) {
      config_set_printer(PRN_ASCII, tmpstr);
    } else if (!strcasecmp(token, "default_bm_printer")) {
      config_set_printer(PRN_BM, tmpstr);
    } else if (!strcasecmp(token, "wm_offset")) {
      if (strcasecmp(tmpstr, "auto")) {
        int x, y;
        
        if (sscanf(tmpstr, "%d , %d", &x, &y) != 2 || x < -50 || x > 50 || y < -50 || y > 50) {
          invalid = TRUE;
          x = y = 0;
        }
        config_set_wmdeco(WM_OFS_X, x);
        config_set_wmdeco(WM_OFS_Y, y);
        config_set_wmdeco(WM_ADJUSTED, WM_ADJ_CONFIG);
      }
    } else if (!strcasecmp(token, "wm_decoration")) {
      if (strcasecmp(tmpstr, "auto")) {
        int lb, rb, tb, bb;
        
        if (sscanf(tmpstr, "%d , %d , %d , %d", &lb, &rb, &tb, &bb) != 4 ||
            lb < 0 || lb > 50 || rb < 0 || rb > 50 || tb < 0 || tb > 50 || bb < 0 || bb > 50) {
          invalid = TRUE;
          lb = rb = bb = 10;
          tb = 30;
        }
        config_set_wmdeco(WM_LB, lb);
        config_set_wmdeco(WM_RB, rb);
        config_set_wmdeco(WM_TB, tb);
        config_set_wmdeco(WM_BB, bb);
        config_set_wmdeco(WM_ADJUSTED, WM_ADJ_CONFIG);
      }
    } else if (config_parse_bool(token, tmpstr, "auto_reset", &tmpval, &invalid)) {
      config_set_auto_reset(tmpval);
    } else if (config_parse_bool(token, tmpstr, "cont_hot_search", &tmpval, &invalid)) {
      config_set_cont_hot_search(tmpval);
    } else if (config_parse_bool(token, tmpstr, "cont_search", &tmpval, &invalid)) {
      config_set_cont_search(tmpval);
    } else if (config_parse_bool(token, tmpstr, "display_on_tv", &tmpval, &invalid)) {
      config_set_tv_disp(tmpval);
    } else if (config_parse_bool(token, tmpstr, "use_interlace", &tmpval, &invalid)) {
      config_set_interlace(tmpval);
    } else if (config_parse_bool(token, tmpstr, "toptext_when_active", &tmpval, &invalid)) {
      config_set_auto_top(tmpval);
    } else if (config_parse_bool(token, tmpstr, "sticky_filechoosers", &tmpval, &invalid)) {
      config_set_sticky_fcs(tmpval);
    } else if (config_parse_bool(token, tmpstr, "spool_only", &tmpval, &invalid)) {
      config_set_spool_only(tmpval);
    } else if (config_parse_int(token, tmpstr, "stationlist_lines", 3, 50, 10, &tmpval,
        &invalid)) {
      config_set_popup_rows(POPUP_SELST, tmpval);
    } else if (config_parse_int(token, tmpstr, "hotlist_lines", 3, 50, 10, &tmpval, &invalid)) {
      config_set_popup_rows(POPUP_HOTLIST, tmpval);
    } else if (config_parse_int(token, tmpstr, "history_lines", 3, 50, 10, &tmpval, &invalid)) {
      config_set_popup_rows(POPUP_HISTORY, tmpval);
    } else if (!invalid) {
      fprintf(stderr, "%s:%d: Warning: Syntax error (ignoring line).\n", fname, lcount);
    }
    if (popup_p != -1) {
      placement_t placement;
      
      for (placement = PLACEMENT_CENTER; placement < PLACEMENT_COUNT; placement++) {
        if (!strcasecmp(placement_str[placement], tmpstr)) {
          config_set_placement(popup_p, placement);
          break;
        }
      }
      if (placement >= PLACEMENT_COUNT) {
        invalid = TRUE;
      }
    } else if (popup_s != -1) {
      if (config_parse_bool(NULL, tmpstr, NULL, &tmpval, &invalid)) {
        config_set_startup(popup_s, tmpval);
      }
    }
    if (invalid) {
      fprintf(stderr, "%s:%d: Warning: Invalid value (ignoring line).\n", fname, lcount);
    }
  }
  return fclose(file);
}


int
config_save(const char *fname) {
  int err;
  FILE *file;
  
  if ((file = fopen(fname, "w"))) {
    fprintf(file, "# This file was created by " VTXWINNAME ", Version " VTXVERSION "\n\n");
    
    fprintf(file, "font=%s\n", vtxfonts[config_get_font()].fontname);
    fprintf(file, "screen_updates=%c%s\n\n",
        tolower(*update_table[config_get_update_speed()].menu_string),
        update_table[config_get_update_speed()].menu_string + 1);
    
    fprintf(file, "history_placement=%s\n", placement_str[config_get_placement(POPUP_HISTORY)]);
    fprintf(file, "toptext_placement=%s\n", placement_str[config_get_placement(POPUP_TOPTEXT)]);
    fprintf(file, "hotlist_placement=%s\n", placement_str[config_get_placement(POPUP_HOTLIST)]);
    fprintf(file, "stationlist_placement=%s\n", placement_str[config_get_placement(POPUP_SELST)]);
    fprintf(file, "history_startup=%s\n", bool_str[config_get_startup(POPUP_HISTORY)]);
    fprintf(file, "toptext_startup=%s\n", bool_str[config_get_startup(POPUP_TOPTEXT)]);
    fprintf(file, "hotlist_startup=%s\n", bool_str[config_get_startup(POPUP_HOTLIST)]);
    fprintf(file, "stationlist_startup=%s\n", bool_str[config_get_startup(POPUP_SELST)]);
    fprintf(file, "toptext_when_active=%s\n", (config_get_auto_top() ? "true" : "false"));
    fprintf(file, "sticky_filechoosers=%s\n\n", (config_get_sticky_fcs() ? "true" : "false"));
    
    fprintf(file, "stationlist_lines=%d\n", config_get_popup_rows(POPUP_SELST));
    fprintf(file, "hotlist_lines=%d\n", config_get_popup_rows(POPUP_HOTLIST));
    fprintf(file, "history_lines=%d\n\n", config_get_popup_rows(POPUP_HISTORY));

    fprintf(file, "display_on_tv=%s\n", (config_get_tv_disp() ? "true" : "false"));
    fprintf(file, "use_interlace=%s\n", (config_get_interlace() ? "true" : "false"));
    fprintf(file, "display_mode=%s\n\n", (config_get_display_mode() == DISPNORM ? "normal" :
        (config_get_display_mode() == DISPTRANS ? "transparent" : "insert")));

    fprintf(file, "auto_reset=%s\n", (config_get_auto_reset() ? "true" : "false"));
    fprintf(file, "cont_hot_search=%s\n", (config_get_cont_hot_search() ? "true" : "false"));
    fprintf(file, "cont_search=%s\n\n", (config_get_cont_search() ? "true" : "false"));

    fprintf(file, "spooldir=%s\n", config_get_spooldir());
    fprintf(file, "spool_only=%s\n", (config_get_spool_only() ? "true" : "false"));
    fprintf(file, "default_cps_printer=%s\n", config_get_printer(PRN_CPS));
    fprintf(file, "default_ps_printer=%s\n", config_get_printer(PRN_PS));
    fprintf(file, "default_ascii_printer=%s\n", config_get_printer(PRN_ASCII));
    fprintf(file, "default_bm_printer=%s\n\n", config_get_printer(PRN_BM));

    if (config_get_wmdeco(WM_ADJUSTED) == WM_ADJ_CONFIG) {
      fprintf(file, "wm_offset=%d,%d\n", config_get_wmdeco(WM_OFS_X), config_get_wmdeco(WM_OFS_Y));
      fprintf(file, "wm_decoration=%d,%d,%d,%d\n", config_get_wmdeco(WM_LB),
          config_get_wmdeco(WM_RB), config_get_wmdeco(WM_TB), config_get_wmdeco(WM_BB));
    } else {
      fprintf(file, "wm_offset=auto\nwm_decoration=auto\n");
    }
  }
  
  if (!file || ferror(file) || fclose(file)) {
    err = errno;
    if (file) {
      fclose(file);
    }
    errno = err;
    return -1;
  }
  return 0;
}


const char*
config_get_printer(printer_t printer) {
  return conf.printer[printer];
}


void
config_set_printer(printer_t printer, const char *str) {
  free(conf.printer[printer]);
  conf.printer[printer] = sstrdup(str);
}


const char*
config_get_spooldir(void) {
  return conf.spooldir;
}


void
config_set_spooldir(const char *str) {
  conf.spooldir = sstrdup(str);
}


int
config_get_tv_disp(void) {
  return conf.tv_disp;
}


void
config_set_tv_disp(int val) {
  conf.tv_disp = !!val;
}


int
config_get_interlace(void) {
  return conf.interlace;
}


void
config_set_interlace(int val) {
  conf.interlace = !!val;
}


int
config_get_sticky_fcs(void) {
  return conf.sticky_fcs;
}


void
config_set_sticky_fcs(int val) {
  conf.sticky_fcs = !!val;
}


int
config_get_auto_top(void) {
  return conf.auto_top;
}


void
config_set_auto_top(int val) {
  conf.auto_top = !!val;
}


int
config_get_auto_reset(void) {
  return conf.auto_reset;
}


void
config_set_auto_reset(int val) {
  conf.auto_reset = !!val;
}


int
config_get_cont_hot_search(void) {
  return conf.cont_hot_search;
}


void
config_set_cont_hot_search(int val) {
  conf.cont_hot_search = !!val;
}


int
config_get_cont_search(void) {
  return conf.cont_search;
}


void
config_set_cont_search(int val) {
  conf.cont_search = !!val;
}


vtxdisp_t
config_get_display_mode(void) {
  return conf.display_mode;
}


void
config_set_display_mode(vtxdisp_t val) {
  conf.display_mode = val;
}


int
config_get_update_speed(void) {
  return conf.update_speed;
}


void
config_set_update_speed(int val) {
  conf.update_speed = val;
}


int
config_get_popup_rows(popup_t popup) {
  return conf.popup_rows[popup];
}


void
config_set_popup_rows(popup_t popup, int val) {
  conf.popup_rows[popup] = val;
}


placement_t
config_get_placement(popup_t popup) {
  return conf.popup_placement[popup];
}


void
config_set_placement(popup_t popup, placement_t val) {
  conf.popup_placement[popup] = val;
}


int
config_get_startup(popup_t popup) {
  return conf.popup_startup[popup];
}


void
config_set_startup(popup_t popup, int val) {
  conf.popup_startup[popup] = !!val;
}


int
config_get_font(void) {
  return conf.font;
}


void
config_set_font(int val) {
  conf.font = val;
}


int
config_get_wmdeco(wmdeco_t wm) {
  return conf.decoration[wm];
}


void
config_set_wmdeco(wmdeco_t wm, int val) {
  conf.decoration[wm] = val;
}


int
config_get_spool_only(void) {
  return conf.spool_only;
}


void
config_set_spool_only(int val) {
  conf.spool_only = !!val;
}
